from __future__ import division

import numbers
import re
import numexpr
from clirender.screen import Screen
from exceptions import NoStretchError

class Layout(object):
	def __init__(self, root):
		self.screen = Screen()
		self.root = root

	def render(self):
		self.screen.clear()

		self.root.render_offset = (0, 0)
		self.root.render_boundary_left_top = [0, 0]
		self.root.render_boundary_right_bottom = list(self.screen.terminal_size)
		self.root.parent = None
		self.root.render_stretch = self.screen.terminal_size[0]

		self.root.render(self)

	def calcRelativeSize(self, size, total, stretch=None, expression=True):
		if not expression:
			# Maybe just absolute integer or float?
			try:
				return float(size)
			except ValueError:
				pass

			# Percent?
			try:
				if size.endswith("%"):
					return float(size[:-1]) / 100 * total
			except ValueError:
				pass

			raise ValueError("Cannot parse %s as non-expression offset/size" % size)

		if not isinstance(size, str) and not isinstance(size, unicode):
			return self.calcRelativeSize(size, total, stretch=stretch, expression=False)


		if stretch is not None:
			stretch = self.calcRelativeSize(stretch, total)


		# Replace percents
		size = re.sub(r"\b([\d\.]+%)(?=\s|$)", lambda x: str(self.calcRelativeSize(x.group(1), total, stretch=stretch, expression=False)), size)

		try:
			data = {}
			if stretch is not None:
				data["stretch"] = stretch

			return numexpr.evaluate(size, local_dict={}, global_dict=data).item()
		except KeyError as e:
			if e.args[0] == "stretch":
				raise NoStretchError()


from node import Node

class Container(Node):
	container = True
	text_container = False

	def __init__(self, type=None, children=[]):
		super(Container, self).__init__(children=children)

		self.type = type

	def render(self, layout, dry_run=False):
		if len(self.children) != 1:
			raise ValueError("<Container> can contain only one node")

		return self.renderChild(
			layout, self.children[0], dry_run=dry_run,

			offset=self.render_offset,
			boundary_left_top=self.render_boundary_left_top,
			boundary_right_bottom=self.render_boundary_right_bottom,
			stretch=self.render_stretch
		)


from rect import Rect

class AlignRight(Rect):
	container = True
	text_container = False

	def __init__(self, width, height=None, bg=None, children=[]):
		super(AlignRight, self).__init__(width=width, height=height, bg=bg, children=children)

	def render(self, layout, dry_run=False):
		if len(self.children) != 1:
			raise ValueError("<AlignRight> can contain only one node")

		# Get child size
		child_width, child_height = self.guessContainerSize(layout)

		# Guess container size
		if self.height is None:
			height = child_height
		else:
			height = self.height

		x1, y1, x2, y2 = super(AlignRight, self).render(layout, dry_run=dry_run, height=height)
		self.renderChildren(layout, x1, y1, x2, y2, dry_run=dry_run, child_width=child_width)

		return x1, y1, x2, y2

	def guessContainerSize(self, layout):
		x1, y1, x2, y2 = super(AlignRight, self).render(layout, dry_run=True, width=self.width, height=self.height or 0)

		return self.renderChildren(layout, x1, y1, x2, y2, dry_run=True)

	def renderChildren(self, layout, x1, y1, x2, y2, dry_run=False, child_width=None):
		if child_width is not None:
			x1 = x2 - child_width

		child = self.children[0]
		child_x1, child_y1, child_x2, child_y2 = self.renderChild(
			layout, child, dry_run=dry_run,

			offset=(x1, y1),

			boundary_left_top=(
				x1,
				y1 if self.height is not None else self.render_boundary_left_top[1]
			),
			boundary_right_bottom=(
				x2,
				y2 if self.height is not None else self.render_boundary_right_bottom[1]
			),

			stretch=self.render_stretch
		)

		return child_x2 - child_x1, child_y2 - child_y1


from rect import Rect
from switch import Switch
from ..exceptions import NoStretchError

class StackPanel(Rect):
	container = True
	text_container = False

	def __init__(self, width=None, height=None, bg=None, orientation="horizontal", hspacing=0, wspacing=0, children=[]):
		super(StackPanel, self).__init__(width=width, height=height, bg=bg, children=children)

		self.vertical = orientation.lower() == "vertical"
		self.hspacing = hspacing
		self.wspacing = wspacing

	def render(self, layout, dry_run=False):
		# Guess container size
		if self.vertical:
			stretch = self.width
		else:
			stretch = self.height

		if self.width is None or self.height is None:
			# Let guessContainerSize() find out 'stretch' if it is not given
			width, height, stretch = self.guessContainerSize(layout, stretch=stretch)
			if self.width is not None:
				width = self.width
			if self.height is not None:
				height = self.height
		else:
			# If the size is given, don't do unnecessary actions
			width, height = self.width, self.height

		x1, y1, x2, y2 = super(StackPanel, self).render(layout, dry_run=dry_run, width=width, height=height)
		self.renderChildren(layout, x1, y1, x2, y2, dry_run=dry_run, stretch=stretch)

		return x1, y1, x2, y2

	def guessContainerSize(self, layout, stretch=None):
		x1, y1, x2, y2 = super(StackPanel, self).render(layout, dry_run=True, width=self.width or 0, height=self.height or 0)

		return self.renderChildren(layout, x1, y1, x2, y2, dry_run=True, stretch=stretch)

	def renderChildren(self, layout, x1, y1, x2, y2, dry_run=False, stretch=None):
		wspacing = layout.calcRelativeSize(self.wspacing, self.render_boundary_right_bottom[0] - self.render_boundary_left_top[0], self.render_stretch)
		hspacing = layout.calcRelativeSize(self.hspacing, self.render_boundary_right_bottom[1] - self.render_boundary_left_top[1], self.render_stretch)

		cur_x, cur_y = x1, y1
		max_width, max_height = 0, 0

		rows_columns = []

		row_column_has_stretch_problems = False
		rows_columns_no_stretch = []

		for child in self.children:
			if isinstance(child, Switch):
				# <Switch /> can be used to switch direction
				if self.vertical:
					if not row_column_has_stretch_problems:
						rows_columns_no_stretch.append(cur_y - y1)

					rows_columns.append((max_width, cur_y - y1))

					cur_x += max_width + wspacing
					cur_y = y1
				else:
					if not row_column_has_stretch_problems:
						rows_columns_no_stretch.append(cur_x - x1)

					rows_columns.append((cur_x - x1, max_height))

					cur_x = x1
					cur_y += max_height + hspacing

				max_width = 0
				max_height = 0

				row_column_has_stretch_problems = False
				continue

			boundary_left_top = self.render_boundary_left_top
			boundary_right_bottom = self.render_boundary_right_bottom
			if self.width is not None:
				boundary_left_top[0] = x1
				boundary_right_bottom[0] = x2
			if self.height is not None:
				boundary_left_top[1] = y1
				boundary_right_bottom[1] = y2

			try:
				child_x1, child_y1, child_x2, child_y2 = self.renderChild(
					layout, child, dry_run=dry_run,

					offset=(cur_x, cur_y),
					boundary_left_top=boundary_left_top,
					boundary_right_bottom=boundary_right_bottom,
					stretch = stretch
				)
			except NoStretchError:
				row_column_has_stretch_problems = True
				continue

			max_width = max(max_width, child_x2 - child_x1)
			max_height = max(max_height, child_y2 - child_y1)

			if self.vertical:
				cur_y = child_y2 + hspacing
			else:
				cur_x = child_x2 + wspacing

		if len(rows_columns_no_stretch) < len(rows_columns):
			# Some rows/columns used 'stretch' variable, which was not calculated yet
			if self.vertical:
				stretch = sum(rows_columns_no_stretch) + cur_y - y1
			else:
				stretch = sum(rows_columns_no_stretch) + cur_x - x1

			return self.renderChildren(layout, x1, y1, x2, y2, dry_run=dry_run, stretch=stretch)

		if self.vertical:
			rows_columns.append((max_width, cur_y - y1))
			return sum(value[0] for value in rows_columns), max(value[1] for value in rows_columns), stretch
		else:
			rows_columns.append((cur_x - x1, max_height))
			return max(value[0] for value in rows_columns), sum(value[1] for value in rows_columns), stretch


import sys, platform
import size

from colorama import Fore, Back, Style

import colorama

class Screen(object):
	class Resized(Exception):
		pass

	def __init__(self):
		if platform.system().lower() == "windows":
			build = int(platform.win32_ver()[1].split(".")[-1])
			if build < 10525:
				# Before build 10525, ANSI escapes were not handled by Windows
				self.color_support = "simulation"

				# Colorama will simulate some ANSI escapes with WinAPI calls
				colorama.init()
			elif build < 16257:
				# Before build 16257, ANSI escapes were enabled by default
				self.color_support = "full"
			elif build >= 16257:
				# As of build 16257, we need to enable ANSI escapes manually
				self.color_support = "full"

				from ctypes import windll, c_int, byref
				stdout_handle = windll.kernel32.GetStdHandle(c_int(-11))
				mode = c_int(0)
				windll.kernel32.GetConsoleMode(c_int(stdout_handle), byref(mode))
				mode = c_int(mode.value | 4)
				windll.kernel32.SetConsoleMode(c_int(stdout_handle), mode)
		else:
			# On Linux and Darwin, ANSI escapes just work.
			self.color_support = "full"

		self.terminal_size = self.getTerminalSize()

	def loop(self, handler):
		while True:
			cur = self.getTerminalSize()
			if self.terminal_size != cur:
				raise Screen.Resized
			self.terminal_size = cur

			handler(self)

	def getTerminalSize(self):
		return size.get_terminal_size()

	def write(self, data):
		sys.stdout.write(data)


	def clear(self):
		self.write("\033[2J\033[1;1f")

	def moveCursor(self, x, y):
		self.write("\x1b[%d;%dH" % (y + 1, x + 1))

	def printAt(self, text, x, y):
		self.moveCursor(int(x), int(y))
		self.write(text)

	def decodeColor(self, color, foreground):
		if "|" in color:
			color, fallback = color.split("|", 1)
		else:
			fallback = None

		color = color.strip()

		if color.lower() in ["black", "blue", "cyan", "green", "magenta", "red", "white", "yellow"]:
			# Colorama supports these names of colors
			return getattr(Fore if foreground else Back, color.upper())
		elif color.startswith("#"):
			# Hex color
			color = color[1:]

			if len(color) == 3:
				# #ABC = #AABBCC
				color = color[0] * 2 + color[1] * 2 + color[2] * 2

			if len(color) != 6:
				raise ValueError("Invalid %s color #%s" % ("foreground" if foreground else "background", color))

			if self.color_support != "full":
				if fallback is not None:
					return self.decodeColor(fallback, foreground)
				raise ValueError("24-bit colors don't work on this platform")

			r, g, b = color[:2], color[2:4], color[4:]
			r, g, b = int(r, 16), int(g, 16), int(b, 16)

			return "\x1B[%s;2;%s;%s;%sm" % (38 if foreground else 48, r, g, b)
		else:
			raise ValueError("Unknown %s color %s" % ("foreground" if foreground else "background", color))

	def colorize(self, text, fg=None, bg=None, bright=False):
		if fg is not None:
			text = self.decodeColor(fg, True) + text
		if bg is not None:
			text = self.decodeColor(bg, False) + text
		if bright:
			text = Style.BRIGHT + text

		text += Style.RESET_ALL
		return text

	def fill(self, x1, y1, x2, y2, char=" ", style=lambda s: s):
		s = char * (int(x2) - int(x1))
		for y in range(int(y1), int(y2)):
			self.printAt(style(s), x=int(x1), y=y)


from generator import Generator
from container import Container

class Range(Generator):
	def __init__(self, from_, to, step=1, slot=None, children=[]):
		super(Range, self).__init__(children=children)

		self.from_ = from_
		self.to = to
		self.step = step
		self.slot = slot

	def generate(self, slots, defines):
		from ..xml_parser import handleElement

		try:
			from_ = int(self.from_)
		except ValueError:
			raise ValueError("'from' attribute of <Range> must be an integer")

		try:
			to = int(self.to)
		except ValueError:
			raise ValueError("'to' attribute of <Range> must be an integer")

		try:
			step = int(self.step)
		except ValueError:
			raise ValueError("'step' attribute of <Range> must be an integer")

		new_slots = dict(**slots)

		res = []
		for i in range(from_, to, step):
			if self.slot is not None:
				new_slots[self.slot] = str(i)

			for child in self.children:
				res += handleElement(child, defines, dict(**new_slots))

		return res


from node import Node

class Switch(Node):
	container = False
	text_container = False

	def __init__(self):
		super(Switch, self).__init__()

	def render(self, layout, dry_run=False):
		raise ValueError("<Switch> must be used inside <StackPanel> only")


from node import Node

class Rect(Node):
	container = False
	text_container = False

	def __init__(self, width, height, bg=None, children=None):
		super(Rect, self).__init__(children=children)

		self.width = width
		self.height = height
		self.bg = bg

	def render(self, layout, dry_run=False, width=None, height=None):
		if width is None:
			width = self.width
		if height is None:
			height = self.height

		width  = layout.calcRelativeSize(width,  self.render_boundary_right_bottom[0] - self.render_boundary_left_top[0], self.render_stretch)
		height = layout.calcRelativeSize(height, self.render_boundary_right_bottom[1] - self.render_boundary_left_top[1], self.render_stretch)

		x1, y1 = map(max, zip(self.render_offset, self.render_boundary_left_top))

		x2, y2 = self.render_offset[0] + width, self.render_offset[1] + height
		x2, y2 = map(min, zip((x2, y2), self.render_boundary_right_bottom))

		bg = self.inherit("bg")

		if bg is not None:
			if not dry_run:
				layout.screen.fill(x1, y1, x2, y2, char=" ", style=lambda s: layout.screen.colorize(s, bg=bg))

		return x1, y1, x2, y2


from node import Node

class Text(Node):
	container = False
	text_container = True

	def __init__(self, value, width=None, bg=None, color=None, bright=False, fill=False):
		super(Text, self).__init__(value=value)

		self.width = width
		self.bg = bg
		self.color = color
		self.bright = bright is not False
		self.fill = fill is not False

	def render(self, layout, dry_run=False):
		if self.width is not None:
			width = self.width
			width = layout.calcRelativeSize(width, self.render_boundary_right_bottom[0] - self.render_boundary_left_top[0], self.render_stretch)
		else:
			width = len(self.value)

		x1, y1 = map(max, zip(self.render_offset, self.render_boundary_left_top))

		x2, y2 = self.render_offset[0] + width, self.render_offset[1] + 1
		x2, y2 = map(min, zip((x2, y2), self.render_boundary_right_bottom))

		bg = self.inherit("bg")
		color = self.inherit("color")
		bright = self.inherit("bright")

		if color is not None:
			if not dry_run:
				if self.fill:
					layout.screen.fill(x1, y1, x2, y2, char=self.value, style=lambda s: layout.screen.colorize(s, bg=bg, fg=color, bright=bright))
				else:
					layout.screen.printAt(layout.screen.colorize(self.value, bg=bg, fg=color, bright=self.bright), x1, y1)
					layout.screen.printAt(layout.screen.colorize(" " * int(width - len(self.value)), bg=bg), x1 + len(self.value), y1)

		return x1, y1, x2, y2


from lxml import etree

import nodes
from slot import Slot

special_slots = {
	"__unset__": None
}

NoDefault = nodes.NoDefault

def fromXml(code):
	parser = etree.XMLParser(recover=True)
	root = etree.fromstring(code, parser=parser)

	define_nodes = root.findall("Define")

	for node in define_nodes:
		if "name" not in node.attrib:
			raise ValueError("<Define> without 'name' attribute")
		elif getattr(nodes, node.tag, None) is not None:
			raise ValueError("Cannot <Define> core node %s" % node.tag)

	define_nodes.sort(key=lambda define: define.attrib["name"])
	for i in range(len(define_nodes) - 1):
		a = define_nodes[i + 0].attrib["name"]
		b = define_nodes[i + 1].attrib["name"]
		if a == b:
			raise ValueError("Non-unique <Define name='%s'>" % a)

	defines = {}
	for node in define_nodes:
		slots = {}
		container = ""

		children = list(node)
		for child in children:
			if child.tag == "Slot" and "define" in child.attrib:
				name = child.attrib["define"]
				if name in special_slots:
					raise ValueError("Defining special slot :%s" % name)

				slots[name] = child.attrib.get("default", NoDefault)
				if slots[name] is NoDefault:
					if ":default" in child.attrib:
						slot_name = child.attrib[":default"]
						slots[name] = slots.get(slot_name, special_slots.get(slot_name, NoDefault))

				if name == "":
					if "container" not in child.attrib:
						raise ValueError("Deprecated: Set container=node or container=text for <Define name='' />")
					elif child.attrib["container"].lower() == "node":
						container = "node"
					elif child.attrib["container"].lower() == "text":
						container = "text"
					else:
						raise ValueError("Only 'node' or 'text' values are accepted for 'container' attribute of <Define name='' />")

		children = filter(lambda child: (child.tag != "Slot" or "define" not in child.attrib) and child.tag != etree.Comment, children)

		if len(children) != 1:
			raise ValueError("<Define> must have exactly one child")
		if children[0].tag == "Range":
			raise ValueError("<Define> must have exactly one child: cannot guarantee that <Range> is always one child")

		defines[node.attrib["name"]] = nodes.fromXml(elem=children[0], slots=slots, defines=defines, name=node.attrib["name"], container=container)

	return handleElement(root, defines, slots={})[0]


def handleElement(node, defines, slots):
	if node.tag == "Define":
		return []
	elif node.tag is etree.Comment:
		return []
	elif node.tag == "Slot":
		if "define" in node.attrib:
			return []

		name = node.attrib.get("name", "")
		if name not in slots:
			raise ValueError("Unknown slot :%s" % name)

		if isinstance(slots[name], str) or isinstance(slots[name], unicode):
			raise ValueError("Unexpected string slot :%s" % name)
		elif slots[name] is NoDefault:
			raise ValueError("Required slot :%s was not passed (from <%s>)" % (name, node.tag))

		return [slots[name]]

	attrs = {}
	inheritable = {}
	for attr, value in node.attrib.items():
		if attr.startswith(":"):
			attr = attr[1:]
			try:
				if slots[value] is NoDefault:
					raise ValueError("Required slot :%s was not passed (from <%s>)" % (value, node.tag))

				value = slots[value]
			except KeyError:
				raise ValueError("Unknown slot :%s" % value)

		if attr.startswith("inherit-"):
			attr = attr[len("inherit-"):]
			inheritable[attr] = value
		else:
			# Escape keywods
			if attr in ("from"):
				attr += "_"

			attrs[attr] = value

	ctor = getattr(nodes, node.tag, None)
	if ctor is None:
		ctor = defines.get(node.tag, None)
	if ctor is None:
		raise ValueError("Unknown node <%s>" % node.tag)

	if issubclass(ctor, nodes.Generator):
		return handleGenerator(node, defines, slots, ctor=ctor, attrs=attrs, inheritable=inheritable)

	result = None
	if ctor.text_container:
		text_inside = getTextInside(node, slots=slots)

		result = ctor(value=text_inside, **attrs)
	elif ctor.container:
		text_inside = getTextInside(node, slots=slots, allow_nodes=True)
		if text_inside.strip() != "":
			raise ValueError("Text inside <%s>" % node.tag)

		children = []
		for child in node:
			children += handleElement(child, defines, slots)

		result = ctor(children=children, **attrs)
	else:
		if len(node) > 0:
			raise ValueError("Nodes inside <%s>" % node.tag)
		elif node.text is not None and node.text.strip() != "":
			raise ValueError("Text inside <%s>" % node.tag)

		result = ctor(**attrs)

	result.inheritable = inheritable
	return [result]

def handleGenerator(node, defines, slots, ctor, attrs, inheritable):
	node = ctor(children=list(node), **attrs)
	node.inheritable = inheritable
	return node.generate(slots, defines)

def getTextInside(node, slots, allow_nodes=False):
	text = ""
	had_nodes = False

	for item in node.xpath("child::node()"):
		if isinstance(item, str) or isinstance(item, unicode):
			text += item

			if allow_nodes is None and text.strip() != "" and had_nodes:
				raise ValueError("Nodes and text inside <%s>" % node.tag)
		elif item.tag is etree.Comment:
			pass
		elif item.tag == "Slot":
			name = item.attrib.get("name", "")
			if name in slots:
				if slots[name] is NoDefault:
					raise ValueError("Required slot :%s was not passed (from <%s>)" % (name, node.tag))
				elif isinstance(slots[name], str) or isinstance(slots[name], unicode):
					text += slots[name]
				elif allow_nodes is None and had_nodes:
					raise ValueError("Nodes and text inside <%s>" % node.tag)
				elif allow_nodes is False:
					raise ValueError("Nodes inside <%s>" % node.tag)
			else:
				raise ValueError("Unknown slot :%s" % name)
		else:
			if allow_nodes is None and text.strip() != "":
				raise ValueError("Nodes and text inside <%s>" % node.tag)
			elif allow_nodes is False:
				raise ValueError("Nodes inside <%s>" % node.tag)

			had_nodes = True

	return text


try:
	import pypandoc
	readme = pypandoc.convert_file("README.md", "rst")
except:
	readme = open("README.md").read()

from setuptools import setup, find_packages
setup(
	name = "clirender",
	version = "0.14",
	description = "CLI rendering engine for Python",
	long_description = readme,
	install_requires = ["colorama", "numexpr", "lxml"],
	packages = find_packages(),
	author = "Ivanq",
	author_email = "imachug@gmail.com",
	url = "https://github.com/imachug/clirender",
	keywords = ["cli", "cli-application", "rendering-2d-graphics"],
	classifiers = []
)


from node import Node

class NoDefault(object):
	pass

def fromXml(elem, slots, name, defines, container):
	class FromXml(Node):
		def __init__(self, **attrs):
			if self.text_container:
				super(FromXml, self).__init__(value=attrs["value"])
				attrs[""] = attrs["value"]
				del attrs["value"]
			elif self.container:
				super(FromXml, self).__init__(children=attrs["children"])

				if len(attrs["children"]) > 1:
					raise ValueError("Too many nodes in <%s>" % name)
				elif len(attrs["children"]) == 1:
					attrs[""] = attrs["children"][0]
					del attrs["children"]
				else:
					raise ValueError("Too less nodes in <%s>" % name)


			self.slots = dict(**slots)
			for attr, value in attrs.items():
				if attr not in self.slots and attr != "":
					raise ValueError("Unexpected slot :%s" % attr)

				self.slots[attr] = value

			for slot in self.slots.values():
				if isinstance(slot, NoDefault):
					raise ValueError("Required slot :%s not passed" % slot)

		def render(self, *args, **kwargs):
			from ..xml_parser import handleElement
			node = handleElement(elem, defines=defines, slots=self.slots)[0]

			node.render_offset = self.render_offset
			node.render_boundary_left_top = self.render_boundary_left_top
			node.render_boundary_right_bottom = self.render_boundary_right_bottom
			node.parent = self.parent
			node.render_stretch = self.render_stretch

			return node.render(*args, **kwargs)

	FromXml.text_container = container == "text"
	FromXml.container = container == "node"

	return FromXml


class Node(object):
	container = False
	text_container = False

	def __init__(self, children=None, value=[], **kwargs):
		self.render_offset = (None, None)
		self.render_boundary_left_top = [None, None]
		self.render_boundary_right_bottom = [None, None]
		self.parent = None
		self.render_stretch = None

		self.inheritable = {}

		self.children = children
		self.value = value

	def render(self, layout, dry_run=False):
		raise NotImplementedError

	def inherit(self, attr):
		node = self

		value = getattr(node, attr, None)
		try:
			while value.strip().lower() == "inherit":
				parent = node.parent
				if parent is None:
					return None

				value = getattr(parent, attr, None)
				if value is None:
					value = parent.inheritable.get(attr, None)

				# Bypass <Container>
				from container import Container
				if value is None and isinstance(parent, Container):
					value = "inherit"

				node = parent
		except AttributeError:
			pass

		try:
			if isinstance(value, Node) and value.strip().lower() == "inherit":
				return value.inherit(attr)
		except AttributeError:
			pass

		return value


	def renderChild(self, layout, child, dry_run, offset, boundary_left_top, boundary_right_bottom, stretch):
		child.render_offset = offset
		child.render_boundary_left_top = boundary_left_top
		child.render_boundary_right_bottom = boundary_right_bottom
		child.render_stretch = stretch
		child.parent = self

		return child.render(layout, dry_run=dry_run)